
#ifdef ENABLE_CAPTURE
#include "capture.h"
#endif

/* YOU MAY NOT EDIT THIS NEXT TWO LINES OF CODE */
#define VERSION "Naken Web - July 26, 2008 - alpha"
#define COPYRIGHT "Copyright 2004-2008 Michael Kohn"
/* THIS IS NOT GPL */

#define MAXCONN 100
#define MAX_USER_THREADS 4 
#define DEFAULT_PORT 5555
#define GC_TIME 30 
#define PASS64_LEN 512
#define DEFAULT_JPEG_QUALITY 80

#define VIDEO_NUM_404 404
#define VIDEO_NUM_400 400
#define BIGGEST_SUPPORTED_FILE 400000
#define BIGGEST_FILE_CHUNK 1024
#define CHUNKS_PER_SEND 8
#define STATE_IDLE 0
#define STATE_HEADERS 1
#define STATE_SEND_FILE 2
#define NEED_HEADER_NO 0
#define NEED_HEADER_YES 1
#define REQUEST_SINGLE 0
#define REQUEST_MULTIPART 1
#define REQUEST_MULTIPART2 2

#define METHOD_GET 0
#define METHOD_POST 1
#define HANDLER_PLUGIN 2
#define HANDLER_CGI 4

#define BUFFER_SIZE 514
#define QUERY_STRING_SIZE 514   // used for plugins only for now

#ifdef WINDOWS
#define SHUT_RDWR SD_BOTH
#include <winsock2.h>
/* int gettimeofday(struct timeval *tv, struct timezone *tz); */
int gettimeofday(struct timeval *tv, int tz);
#endif

#define FOUR_OH_FOUR "<html><body><br><br><h1>404 - Naken Web FILE NOT FOUND</h1></body></html>\r\n"
#define FOUR_OH_ONE "<html><body><br><br><h1>401 - Naken Web Authorization Required</h1></body></html>\r\n"
#define VIDEO_ERROR "<html><body><br><br><h1>Naken Web Error: AVI frame too big or parse error</h1></body></html>\r\n"
#ifdef ENABLE_CGI
#define FOUR_OH_OH "<html><body><br><br><h1>400 - Naken Web Bad Request</h1></body></html>\r\n"
#endif

struct User
{
  unsigned char location[64];
  char outString[BUFFER_SIZE];
  unsigned char myString[BUFFER_SIZE];
  int ID,r;
  int logontime,idletime;
  char inuse;
  int socketid;
  int sptr,slen;
  int buffer_ptr;
  int curr_frame;
  // FILE *in,*socketfd;
  int in;
#ifdef WITH_MMAP
  long offset;
#endif
  int video_num;
  int mime_type;
  long content_length;
  time_t last_modified;
  int state;
  int need_header;
  int request_type;
  int last_frame;
  unsigned int flags;
  int frame_rate;
  int method;
#ifdef ENABLE_PLUGINS
  char querystring[QUERY_STRING_SIZE];
  struct plugin_t *plugin;
#endif
#ifdef ENABLE_CAPTURE
  unsigned char *jpeg;
  int jpeg_len;
  int jpeg_quality;
  int jpeg_ptr;
#endif
};

// FIXME - a linked list here could save memory
struct video_t
{
  char *filename;
  long *index;
  int fps;
  int total_frames;
  struct timeval tv_start;
  struct timeval tv_total;
#ifdef ENABLE_CAPTURE
  struct capture_info_t *capture_info;
#endif
#ifdef WITH_MMAP
  long file_len;
  unsigned char *mem;
#ifndef WINDOWS
  int fd;
#else
  HANDLE fd;
  HANDLE mem_handle;
#endif
#endif
};

struct alias_t
{
  struct alias_t *next_alias;
  int port;
  int port_start;
  int type;
  int url_len;
  char *url;
  char *port_param;
  char *size_param;
  char *comp_param;
  char *fps_param;
};

#ifdef ENABLE_CGI
struct cgi_handler_t
{
  struct cgi_handler_t *next_handler;
  char *extension;
  char *application;
};
#endif


#ifdef ENABLE_PLUGINS
typedef int (*naken_plugin_init_t)();
typedef int (*naken_plugin_get_t)(int,char*);
typedef int (*naken_plugin_post_t)(int,char*,int);
#ifdef WINDOWS
#define dlsym GetProcAddress
#define dlopen(libname,x) LoadLibrary(libname)
#define dlclose(libhandle) FreeLibrary(libhandle)
#endif

struct plugin_t
{
  struct plugin_t *next_plugin;
  char *alias;
  int alias_len;
#ifndef WINDOWS
  void *dlhandle;
  naken_plugin_get_t naken_plugin_get;
  naken_plugin_post_t naken_plugin_post;
/*
  void naken_plugin_init();
  int naken_plugin_get(int fd_out, char *querystring);
  int naken_plugin_post(int fd_out, char *querystring, int fd_in, int content_length);
*/
#else
  HMODULE dlhandle;
  naken_plugin_init_t naken_plugin_init;
  naken_plugin_get_t naken_plugin_get;
  naken_plugin_post_t naken_plugin_post;
#endif
};
#endif

/* TODO -- Kill these global variables */
extern int sockfd,port;
extern char *runas_user;
extern char *runas_group;
extern int video_count;
extern struct User **users;
extern struct video_t video[100];
extern int debug;
extern unsigned int uptime;
extern unsigned short int serverflags;
/* extern char *config_dir; */
extern char *htdocs_dir;
extern int maxconn;
extern int minconn;
extern int max_idle_time;
extern char user_pass_64[PASS64_LEN];
extern int jpeg_quality;
extern struct alias_t *alias; 
#ifdef ENABLE_CGI
extern struct cgi_handler_t *cgi_handler; 
#endif
#ifdef ENABLE_PLUGINS
extern struct plugin_t *plugin; 
#endif
extern int frame_rate;
extern const char *mime_types[];
extern const char *extensions[];
extern const int mime_type_code[];

/*

User Flags

0: authentication okay
1: did authenication once

*/

