#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef WINDOWS
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <netdb.h>
#include <pthread.h>
#endif
#include <errno.h>
#include <time.h>
#include <string.h>
#ifdef WINDOWS
#include <windows.h>
#include <process.h>
#include <winsock.h>
#endif

#ifdef JPEG_LIB
#include "jpeg_compress.h"
#endif

#ifdef ENABLE_CAPTURE
#include "capture.h"
#endif

#include "globals.h"
#include "functions.h"
#include "fileio.h"
#include "mime_types.h"

/*

This code is copyright 2004-2007 by Michael Kohn
You may not use any of this program in your own programs
without permission from Michael Kohn <mike@mikekohn.net>

*/

/* TODO - Kill these globals */
int sockfd,port;
int video_count;
char *runas_user=0;
char *runas_group=0;
struct User **users;
struct User nulluser;
struct video_t video[100];
int debug;
unsigned int uptime;
unsigned short int serverflags;
/* char *config_dir=0; */
char *htdocs_dir=0;
char user_pass_64[PASS64_LEN];
int maxconn;
int minconn;
int max_idle_time=0;
int jpeg_quality;
struct alias_t *alias;
#ifdef ENABLE_CGI
struct cgi_handler_t *cgi_handler;
#endif
#ifdef ENABLE_PLUGINS
struct plugin_t *plugin;
#endif
int frame_rate;

int file_close(int ID)
{
int in;

  if (users[ID]->in==-1) return 0;

  in=users[ID]->in;
  users[ID]->in=-1;

#ifdef ENABLE_CGI
  if ((users[ID]->mime_type&32768)!=0)
  { return pclose(in); }
    else
#endif
  { return close(in); }
}

int open_file(int ID, char *filename)
{
struct alias_t *curr_alias;
struct stat file_stat;
char full_file[2048];
char *querystring;
#ifdef ENABLE_CGI
char handler_file[2048];
struct cgi_handler_t *curr_handler;
int t;
#endif
#ifdef ENABLE_PLUGINS
struct plugin_t *curr_plugin;
#endif

  if (filename[0]==0) return VIDEO_NUM_404;

#ifdef DEBUG
if (debug==1)
{ printf("filename: %s\n",filename); }
#endif

  curr_alias=alias;

  while (curr_alias!=0)
  {
    if (strncmp(filename,curr_alias->url,curr_alias->url_len)==0)
    {
      users[ID]->video_num=0;

      if (curr_alias->type!=0)
      { users[ID]->request_type=REQUEST_MULTIPART; }

      parse_querystring(ID,filename,curr_alias);
      return users[ID]->video_num;
    }

    curr_alias=curr_alias->next_alias;
  }

#ifdef ENABLE_PLUGINS
  curr_plugin=plugin;

  while(curr_plugin!=0)
  {
    if (strncmp(filename,curr_plugin->alias,curr_plugin->alias_len)==0)
    {
#ifdef DEBUG
printf("Opening plugin\n");
#endif
      // send_header_plugin(ID);
      querystring=filename+curr_plugin->alias_len;
      if (querystring[0]=='?') querystring=querystring+1;
      users[ID]->plugin=curr_plugin;
      strncpy(users[ID]->querystring,querystring,QUERY_STRING_SIZE-1);
/*
      if (curr_plugin->naken_plugin_get(users[ID]->socketid,querystring)!=0)
      { kill_user(ID); }
*/
      return -3;
    }

    curr_plugin=curr_plugin->next_plugin;
  }
#endif

  querystring=get_querystring(filename);

  if (check_valid_file(filename)==-1)
  {
    users[ID]->video_num=-1;
    return VIDEO_NUM_404;
  }

  if (strlen(filename)+strlen(htdocs_dir)+1>=2046)
  { return VIDEO_NUM_404; }

  if (filename[0]=='/' && filename[1]==0)
  {
    sprintf(full_file,"%s/index.html",htdocs_dir);
    users[ID]->mime_type=1;
  }
    else
  {
    url_decode((unsigned char *)filename);
#ifdef DEBUG
printf("url_decode(filename)=%s\n",filename);
#endif

/* FIXME - convert urlencoded stuff */
    if (filename[0]!='/')
    { sprintf(full_file,"%s/%s",htdocs_dir,filename); }
      else
    { sprintf(full_file,"%s%s",htdocs_dir,filename); }

    users[ID]->mime_type=get_mime_type_code(filename);

    if (users[ID]->in!=-1)
    { file_close(ID); }

#ifdef ENABLE_CGI
    setenv("QUERY_STRING",querystring,1);

    if ((users[ID]->mime_type&32768)!=0)
    {
      t=32768;
      curr_handler=cgi_handler;
      while(t!=users[ID]->mime_type)
      {
        curr_handler=curr_handler->next_handler;
        t++;
      }

      if (curr_handler->application!=0)
      {
        sprintf(handler_file,"%s '%s'",curr_handler->application,full_file);
#ifdef DEBUG
printf("handler_file=%s\n",handler_file);
#endif
        // FIXME - This is wrong.. i'll fix it soon
        users[ID]->in=popen(handler_file,"r");
      }
        else
      {
#ifdef DEBUG
printf("full_file=%s\n",full_file);
#endif
        sprintf(handler_file,"'%s'",full_file);
        // FIXME - This is wrong.. i'll fix it soon
        users[ID]->in=popen(handler_file,"r");
      }

      /* unsetenv might not be needed actually */
      unsetenv("QUERY_STRING");

      if (users[ID]->in==-1)
      { return VIDEO_NUM_400; }

      return -2;
    }
#endif
  }

#ifdef DEBUG
if (debug==1)
{ printf("mime-type: %s\n",mime_types[users[ID]->mime_type]); }

if (debug==1)
{ printf("opening: %s\n",full_file); }
#endif

  // users[ID]->in=fopen(full_file,"rb");
#ifndef WINDOWS
  users[ID]->in=open(full_file,O_RDONLY);
#else
  users[ID]->in=open(full_file,O_RDONLY|_O_BINARY);
#endif

  if (users[ID]->in==-1) return VIDEO_NUM_404;

  users[ID]->last_modified=time(NULL);

  /* file len */
  // fseek(users[ID]->in,0,SEEK_END);
/*
  lseek(users[ID]->in,0,SEEK_END);
  users[ID]->content_length=ftell(users[ID]->in);
*/
  // rewind(users[ID]->in);

  fstat(users[ID]->in,&file_stat);
  users[ID]->content_length=file_stat.st_size;

  return -2;
}

int init(int socketid, struct sockaddr_in *cli_addr, int clilen)
{
int k,r,ID;
/* struct hostent *my_hostent=NULL; */
unsigned char *address=0;
char *fullmessage={ "Naken Web Server is full\r\n" };

/*
  if ((serverflags&2)==2)
  { my_hostent=gethostbyaddr((char *)&(cli_addr->sin_addr.s_addr),4,AF_INET); }
*/

  for (r=0; r<maxconn; r++)
  { 
    if (users[r]->inuse==0) 
    { break; } 
  }

  if (r>=maxconn)
  {
#ifdef DEBUG
    if (debug==1) printf("Naken Web server full\n");
#endif
    send_data(socketid,fullmessage,strlen(fullmessage)); 
    socketdie(socketid);
    return -1;
  }

  if (users[r]->idletime==-1)
  { 
#ifdef DEBUG
    if (debug==1) printf("Allocing new user %d\n",r);
#endif
    users[r]=malloc(sizeof(struct User)); 
  }

#ifdef DEBUG
  if (debug==1)
  {
    printf("searched for open User: found %d\n",r);
    fflush(stdout);
  }
#endif

/*
  if (my_hostent!=NULL && (serverflags&2)==2)
  { safestrncpy(users[r]->location,my_hostent->h_name,61); }
    else
*/
  { 
    address=(unsigned char*)&cli_addr->sin_addr.s_addr;
    sprintf((char *)users[r]->location,"%d.%d.%d.%d",address[0],address[1],address[2],address[3]);
  }

#ifdef DEBUG
  if (debug==1)
  {
    printf("Connected from location: %s\n",users[r]->location);
    fflush(stdout);
  }
#endif

  k=0;
  if ((serverflags&8)==8)
  {
    for (k=0; k<maxconn; k++)
    {
      if (users[k]->inuse!=1) continue;
      if (strcmp((char *)users[r]->location,(char *)users[k]->location)==0)
      { k=-1; break; }
    }
  }

  ID=r;
  // FIXME - can we memset this or is faster to do this?
  users[ID]->inuse=200;
  users[ID]->ID=r;
  users[ID]->socketid=socketid;

  users[ID]->slen=0;
  users[ID]->sptr=0;
  users[ID]->buffer_ptr=0;
  users[ID]->logontime=time(NULL);
  users[ID]->idletime=time(NULL);
  users[ID]->curr_frame=-1;
  users[ID]->video_num=-1;
  users[ID]->in=-1;
  users[ID]->inuse=1;
  users[ID]->state=STATE_IDLE;
#ifdef ENABLE_PLUGINS
  users[ID]->plugin=0;
#endif

/*
  users[ID]->socketfd=fdopen(socketid,"rb+");

  if (users[ID]->socketfd==0) printf("Problem opening file desc.\n\n");
*/

#ifndef WINDOWS
  fcntl(users[ID]->socketid, F_SETFL, O_NONBLOCK);
#else
  u_long arg=1L;
  ioctlsocket(users[ID]->socketid, FIONBIO, (u_long FAR *)&arg);
#endif

/* hmmm 

  if (setsockopt(sockfd,SOL_SOCKET,SO_RCVLOWAT,&sopt,sizeof(sopt)))
  { printf("socket options SO_RCVLOWAT can't be set.\n"); }
  if (setsockopt(sockfd,SOL_SOCKET,SO_SNDLOWAT,&sopt,sizeof(sopt)))
  { printf("socket options SO_SNDLOWAT can't be set.\n"); }
*/

#ifdef DEBUG
if (debug==1)
{
  printf("ID=%d ready to go\n",ID);
  fflush(stdout);
}
#endif

  return 0;
}

void kill_user(int ID)
{

/* printf("kill user  inuse=%d\n",users[ID]->inuse); */

#ifdef DEBUG
  if (debug==1)
  {
    printf("Getting ready to close socket.\n");
    if (users[ID]->inuse!=4)
      printf("Socket Close Return value: %d:\n",socketdie(users[ID]->socketid));
  }
    else
#endif
  { if (users[ID]->inuse!=4) socketdie(users[ID]->socketid); }

  users[ID]->socketid=0;

#ifdef DEBUG
  if (debug==1) printf("Connection dead: %d\n",ID);
  fflush(stdout);
#endif

  if (users[ID]->in!=-1)
  {
    /* fclose(users[ID]->in); */
    file_close(ID);
  }

  users[ID]->inuse=0;
  users[ID]->idletime=time(NULL);
}

int calc_frame(int ID)
{
unsigned char temp[8];
struct timeval tv_now;
int frame;
int t,r;
#ifndef WITH_MMAP
// int l;
#endif

  /* gettimeofday(&tv_now,DST_NONE); */
  gettimeofday(&tv_now,0);

  t=((tv_now.tv_sec-video[users[ID]->video_num].tv_start.tv_sec)*1000)+
    ((tv_now.tv_usec-video[users[ID]->video_num].tv_start.tv_usec)/1000);

  r=(int)(((double)t/1000)*(double)video[users[ID]->video_num].fps);
  t=r/video[users[ID]->video_num].total_frames;
  r=r%video[users[ID]->video_num].total_frames;

  if (t>1000 || r<0)
  {
    if (r<0) r=0;
    gettimeofday(&video[users[ID]->video_num].tv_start,0);
    t=r*(1000000/video[users[ID]->video_num].fps);
    video[users[ID]->video_num].tv_start.tv_sec+=(t/1000000);
    video[users[ID]->video_num].tv_start.tv_usec=(t%1000000);

#ifdef DEBUG
if (debug==1)
{
  printf("video flipped\n");
}
#endif
  }

  frame=r;

#ifdef WITH_MMAP
  users[ID]->offset=video[users[ID]->video_num].index[r]+4;
/*
  if (l>video[users[ID]->video_num].index[r])
  { users[ID]->offset=video[users[ID]->video_num].index[r]; }
    else
  { users[ID]->offset=video[users[ID]->video_num].index[r]-l; }
*/

  users[ID]->content_length=video[users[ID]->video_num].mem[users[ID]->offset]+
                     (video[users[ID]->video_num].mem[users[ID]->offset+1]<<8)+
                     (video[users[ID]->video_num].mem[users[ID]->offset+2]<<16)+
                     (video[users[ID]->video_num].mem[users[ID]->offset+3]<<24);
  users[ID]->offset+=4;
#else
  // FIXME - can we seek by offsets forward and does it give a speed increase?
  // l=ftell(users[ID]->in);

/*
  if (l>video[users[ID]->video_num].index[r])
  { fseek(users[ID]->in,video[users[ID]->video_num].index[r],SEEK_SET); }
    else
  { fseek(users[ID]->in,video[users[ID]->video_num].index[r]-l,SEEK_CUR); }
*/
/*
  if (l>video[users[ID]->video_num].index[r])
  { lseek(users[ID]->in,video[users[ID]->video_num].index[r],SEEK_SET); }
    else
  { lseek(users[ID]->in,video[users[ID]->video_num].index[r]-l,SEEK_CUR); }
*/

  lseek(users[ID]->in,video[users[ID]->video_num].index[r],SEEK_SET);

  // read_int(users[ID]->in);
  // users[ID]->content_length=read_int(users[ID]->in);
  read(users[ID]->in,temp,8);
  users[ID]->content_length=temp[4]+((temp[5])<<8)+
                            ((temp[6])<<16)+((temp[7])<<24);
#endif

  if (users[ID]->content_length>BIGGEST_SUPPORTED_FILE || users[ID]->content_length==0)
  {
    send_video_error(ID);
    kill_user(ID);
    return -1;
  }

  users[ID]->last_modified=time(NULL);
  users[ID]->mime_type=4; /* JPEG DUH */

  return frame;
}

void chat(int *params)
{
int t=0,r,trim;
int ID=0;
int msock=0;
fd_set readset;
struct timeval tv;
char *outString;
int gc_time,thread_num;
int dirty_buffer;
char *command,*param;

  thread_num=params[0];
  gc_time=time(NULL);

  while(1)
  {
    if (time(NULL)-gc_time>GC_TIME)
    { 
      for (r=minconn+thread_num; r<maxconn; r=r+MAX_USER_THREADS)
      {
        if (users[r]->inuse==0 && users[r]->idletime!=-1)
        {
#ifdef DEBUG
          if (debug==1) printf("Deallocating line %d\n",r);
#endif
          if (time(NULL)-users[r]->idletime>GC_TIME)
          free(users[r]);
          users[r]=&nulluser;
        }
      }

      if (max_idle_time>0)
      {
        for (r=thread_num; r<maxconn; r=r+MAX_USER_THREADS)
        {
          if ((users[r]->inuse!=0) && time(NULL)-users[r]->idletime>max_idle_time)
          {
            kill_user(r);
          }
        }
      }

      gc_time=time(NULL);
    }

    FD_ZERO(&readset);
    msock=0;
    dirty_buffer=0;

    for (r=thread_num; r<maxconn; r=r+MAX_USER_THREADS)
    {
      if (users[r]->inuse==1)
      {
        FD_SET(users[r]->socketid,&readset);
        if (msock<users[r]->socketid) msock=users[r]->socketid;
        if (users[r]->sptr<users[r]->slen || users[r]->state==STATE_SEND_FILE)
        { dirty_buffer=1; }
      }
    }

    if (dirty_buffer==0)
    {    
      tv.tv_sec=1;
      tv.tv_usec=0;
    }
      else
    {
      tv.tv_sec=0;
      tv.tv_usec=1;
    }

#ifdef WINDOWS
    if (msock==0)
    {
      _sleep(1000);
      continue;
    }
#endif

    if ((t=select(msock+1,&readset,NULL,NULL,&tv))==-1)
    {
#ifdef WINDOWS
      if (WSAGetLastError()!=WSANOTINITIALISED) printf("yes %d\n",errno);
      if (WSAGetLastError()!=WSAEINTR)
#else
      if (errno!=EINTR)
#endif
      { 
#ifdef DEBUG
        if (debug==1) printf("not EINTR %d\n",thread_num);
#endif
      }
         else
      { 
#ifdef DEBUG
        if (debug==1) printf("Problem with select\n"); 
#endif
        continue;
      }
    }

    for (ID=thread_num; ID<maxconn; ID=ID+MAX_USER_THREADS)
    { 
      errno=0;
      if (users[ID]==0) continue;
      if (users[ID]->inuse!=1) continue;

/*
printf("Checking line %d     %d %d %d\n",ID,users[ID]->sptr,users[ID]->slen,users[ID]->ptr);
*/
/* printf("state=%d  need_header=%d (%d)\n",users[ID]->state,users[ID]->need_header,ID); */

      if (users[ID]->state==STATE_SEND_FILE)
      {
        /* Do nothing */
      }
        else
      if (users[ID]->sptr<users[ID]->slen || FD_ISSET(users[ID]->socketid,&readset))
      {
        r=buffered_read(ID);
        if (r<0) { kill_user(ID); continue; }
        if (r==0) continue;

        if (user_pass_64[0]!=0)
        {
          if (strncmp(users[ID]->outString,"Authorization: Basic ",sizeof("Authorization: Basic ")-1)==0)
          {
            if ((users[ID]->flags&2)==0 &&
                strcmp(user_pass_64,users[ID]->outString+sizeof("Authorization: Basic ")-1)==0)
            {
              users[ID]->flags|=1;
            }
            users[ID]->flags|=2;
          }
        }

        if (r==2) users[ID]->state=STATE_SEND_FILE;
      }
        else
      { continue; }

      if (users[ID]->state==STATE_SEND_FILE)
      {
        if (user_pass_64[0]!=0)
        {
          if ((users[ID]->flags&1)!=1)
          {
            send_401(ID);
            continue;
          }
        }

#ifdef ENABLE_PLUGINS
        if (users[ID]->video_num==-3)  // PLUGIN
        {
           if (users[ID]->method==METHOD_GET)
           {
             send_header_plugin(ID);
             if (users[ID]->plugin->naken_plugin_get(users[ID]->socketid,users[ID]->querystring)!=0)
             {
               kill_user(ID);
               continue;
             }
           }
             else
           // this is totally fuckered.. need to give content length
           if (users[ID]->method==METHOD_POST)
           {
             send_header_plugin(ID);
             if (users[ID]->plugin->naken_plugin_post(users[ID]->socketid,users[ID]->querystring,0)!=0)
             {
               kill_user(ID);
               continue;
             }
           }

           users[ID]->state=STATE_IDLE;
           users[ID]->plugin=0;
          
           continue;
        }
          else
#endif
        if (users[ID]->video_num==-2)  // FILE OR CGI
        {
          send_file(ID);
          continue;
        }
          else
        if (users[ID]->video_num>=video_count)
        {
          send_error(ID,"404 Not Found",FOUR_OH_FOUR,sizeof(FOUR_OH_FOUR));
          continue;
        }
          else
        if (users[ID]->video_num>=0)
        {

#ifdef ENABLE_CAPTURE
          if (video[users[ID]->video_num].capture_info!=0)
          {
            send_capture_frame(ID);
            continue;
          }
#endif

          if (users[ID]->in==-1)
          {
#ifndef WINDOWS
            users[ID]->in=open(video[users[ID]->video_num].filename,O_RDONLY);
#else
            users[ID]->in=open(video[users[ID]->video_num].filename,O_RDONLY|_O_BINARY);
#endif

            if (users[ID]->in==-1)
            {
#ifdef ENABLE_CGI
              if ((users[ID]->mime_type&32768)==0)
              { send_error(ID,"404 Not Found",FOUR_OH_FOUR,sizeof(FOUR_OH_FOUR)); }
                else
              { send_error(ID,"400 Bad Request",FOUR_OH_OH,sizeof(FOUR_OH_OH)); }
#else
              send_error(ID,"404 Not Found",FOUR_OH_FOUR,sizeof(FOUR_OH_FOUR));
#endif
              continue;
            }
          }

          if (users[ID]->need_header!=NEED_HEADER_NO)
          {
            users[ID]->idletime=time(NULL);
            r=calc_frame(ID);

            if (r==users[ID]->last_frame) continue;

            if (abs(r-users[ID]->last_frame)<
                (video[users[ID]->video_num].fps/
                users[ID]->frame_rate)) continue; // FIXME

#ifdef DEBUG
if (debug==1)
{
  printf("Frame %d/%d  camera=%d\n",r,video[users[ID]->video_num].total_frames,users[ID]->video_num);
}
#endif
            users[ID]->last_frame=r;
          }

          send_file(ID);

          continue;
        }
          else
        { continue; } //FIXME - why?
      }

      // if (users[ID]->video_num>=0) continue;
      // if (users[ID]->video_num!=-1) continue;
      if (users[ID]->state!=STATE_IDLE) continue;

      outString=users[ID]->outString;

      trim=0;

      while(outString[trim]==' ') trim++;

      users[ID]->idletime=time(NULL);

#ifdef DEBUG
      if (debug==1)
      {
        printf("Read in: %d bytes on thread %d.\n",(int)strlen(outString),thread_num);
        printf("%d typed: %s\n",ID,outString);
        fflush(stdout); 
      }
#endif

      r=0;
      while(outString[r]==' ' && outString[r]!=0) r++;

      command=outString+r;
      r=0;
      while(command[r]!=' ' && command[r]!='\r' && command[r]!='\n' && command[r]!=0) r++;
      command[r++]=0;

#ifdef DEBUG
      if (debug==1)
      {
        printf("command=%s\n",command);
      }
#endif

      param=command+r;
      r=strlen(param)-1;
      while((param[r]==' ' || param[r]=='\r') || param[r]=='\n') param[r--]=0;

      if (strcasecmp(command,"get")==0)
      {
        users[ID]->state=STATE_HEADERS;
        users[ID]->method=METHOD_GET;

        r=conv_num(param+1);
        if (r!=users[ID]->video_num && users[ID]->in!=-1)
        {
          file_close(ID);
        }

        users[ID]->video_num=r;
        users[ID]->need_header=NEED_HEADER_YES;
        users[ID]->request_type=REQUEST_SINGLE;
        users[ID]->last_frame=-1;
        users[ID]->flags=0;
        users[ID]->frame_rate=frame_rate;
#ifdef ENABLE_CAPTURE
        users[ID]->jpeg_quality=jpeg_quality;
#endif

        if (users[ID]->video_num==-1)
        {
          r=1;
          while(param[r]!=' ' && param[r]!=0) r++;
          param[r]=0;
          users[ID]->video_num=open_file(ID,param);

#ifdef DEBUG
if (debug==1)
{ printf("video_num=%d\n",users[ID]->video_num); }
#endif

        }
      }
#ifdef ENABLE_CGI
        else
      if (strcasecmp(command,"post")==0)
      {
         // complete me
         // this maybe should go above in the GET section
      }
#endif
        else
      {
        /* Unknown command */
        kill_user(ID);
      }
    }
  }
}

int main(int argc,char *argv[])
{
struct sockaddr_in serv_addr;
int r,sopt=1;
struct linger mylinger;
int newsockfd;
#ifndef WINDOWS
socklen_t clilen;
#else
int clilen;
#endif
struct sockaddr_in cli_addr;
int threads[MAX_USER_THREADS];
#ifndef WINDOWS
pthread_t pid;
#endif
#ifdef WINDOWS
WSADATA wsaData;
WORD wVersionRequested=MAKEWORD(1,1);
int Win32isStupid;

  Win32isStupid=WSAStartup(wVersionRequested, &wsaData);
  if (Win32isStupid)  /* Shouldn't this always be true? :) */
  {
    printf("Winsock can't start.\n");
    exit(1);
  }

/*
  printf("Max sockets: %d\n",wsaData.iMaxSockets);
  printf("Description: %s\n",wsaData.szDescription);
*/
#endif

  maxconn=MAXCONN;
  minconn=10;
  max_idle_time=60;
  video_count=0;
  setup(argc,argv);

#ifndef WINDOWS
  if (debug==0)
  {
    r=fork();

    if (r==-1)
    { perror("Error: Could not fork\n"); exit(3); }
      else
    if (r==0)
    {
      close(STDIN_FILENO);
      close(STDERR_FILENO);

      if (setsid()==-1)
      { exit(4); }
    }
      else
    { return 0; }
  }
#endif

#ifdef DEBUG
  if (debug==1)
  {
    printf("minconn: %d\n",minconn);
    printf("maxconn: %d\n",maxconn);
  }
#endif

  if ((sockfd=socket(AF_INET,SOCK_STREAM,0))<0)
  { printf("Can't open socket.\n"); exit(1); }

  /* bzero((char*)&serv_addr,sizeof(serv_addr)); */
  memset((char*)&serv_addr,0,sizeof(serv_addr));
  serv_addr.sin_family=AF_INET;
  serv_addr.sin_addr.s_addr=htonl(INADDR_ANY);
  serv_addr.sin_port=htons(port);

  mylinger.l_onoff=0;
  mylinger.l_linger=1;

#ifndef WINDOWS
  if (setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,&sopt,sizeof(sopt)))
  { printf("socket options REUSERADDR can't be set.\n"); exit(1); }
  if (setsockopt(sockfd,SOL_SOCKET,SO_LINGER,&mylinger,sizeof(struct linger)))
  { printf("socket options LINGER can't be set.\n"); exit(1); }
#else
  if (setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,(char *)&sopt,sizeof(sopt)))
  { printf("socket options REUSERADDR can't be set.\n"); exit(1); }
  if (setsockopt(sockfd,SOL_SOCKET,SO_LINGER,(char *)&mylinger,sizeof(struct linger)))
  { printf("socket options LINGER can't be set.\n"); exit(1); }

  timeBeginPeriod(5);
#endif


  if(bind(sockfd,(struct sockaddr *)&serv_addr,sizeof(serv_addr))<0)
  { printf("server can't bind.\n"); exit(1); }

  listen(sockfd,5);
#ifndef WINDOWS
  setsignals();
  set_runas();
#endif

/* YOU MAY NOT EDIT THE NEXT LINE OF CODE */
  printf("\n"VERSION"\n"COPYRIGHT"\n\n");
/* NAKEN CHAT IS NOT GPL */

  printf("Started on port: %d\n",port);
  fflush(stdout);

  if (debug==0) close(STDOUT_FILENO);

  nulluser.inuse=0;
  nulluser.idletime=-1;
  users=malloc(sizeof(char *)*maxconn);

  for (r=0; r<maxconn; r++)
  {
    users[r]=&nulluser;
  }

  for (r=0; r<minconn; r++)
  {
    users[r]=malloc(sizeof(struct User));
    users[r]->inuse=0;
  }

  for (r=0; r<MAX_USER_THREADS; r++)
  {
    threads[r]=r;
#ifndef WINDOWS
    pthread_create(&pid,NULL,(void*)chat,&threads[r]);
#else
    _beginthread((void*)chat,0,&threads[r]);
#endif
  }

  clilen=sizeof(cli_addr);

  while(1)
  {
    newsockfd=accept(sockfd,(struct sockaddr *)&cli_addr,&clilen);

    if (newsockfd<0)
    {
#ifdef DEBUG
      if (debug==1) { printf("server accept error.\n"); }
#endif
    }
      else
    { 
#ifdef DEBUG
      if (debug==1)
      {
        printf("New socket accepted\n");
        fflush(stdout);
      }
#endif
      init(newsockfd,&cli_addr,clilen);
    }
  }

  return 0;
}

