#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#ifndef WINDOWS
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <netdb.h>
#include <pthread.h>
#endif
#include <errno.h>
#include <time.h>
#include <string.h>
#ifdef WINDOWS
#include <windows.h>
#include <winsock.h>
#endif

#include "globals.h"
#include "functions.h"

/*

This code is copyright 2004-2007 by Michael Kohn
You may not use any of this program in your own programs
without permission from Michael Kohn <mike@mikekohn.net>

*/

int send_data(int socketid, char *message, int message_len)
{
struct timeval tv;
fd_set writeset;
int t;

  FD_ZERO(&writeset);
  FD_SET(socketid,&writeset);

  // FIXME - Time out of null is bad bad bad

  tv.tv_sec=1;
  tv.tv_usec=0;

  if (((t=select(socketid+1,NULL,&writeset,NULL,NULL))==-1) && (errno!=EINTR))
  {
    printf("Problem with select\n");
    return -1;
  }
    else
  {
    if (t==0) return -1;
    t=send(socketid,message,message_len,0);
#ifdef WINDOWS
    _commit(socketid);
#else
    fsync(socketid);
#endif
    return t;
  }

  return 0;
}

int send_file(int ID)
{
char temp_string[BIGGEST_FILE_CHUNK];
int t,k,r,c;

  if (users[ID]->in==-1)
  {
    kill_user(ID);
    return -1;
  }

  if (users[ID]->need_header==NEED_HEADER_YES)
  {
    if (users[ID]->request_type==REQUEST_SINGLE)
    {
#ifdef ENABLE_CGI
      if ((users[ID]->mime_type&32768)!=0)
      { send_header_cgi(ID); }
        else
#endif
      { send_header(ID); }
    }
      else
    if (users[ID]->request_type==REQUEST_MULTIPART)
    { send_header_multipart(ID); }
      else
    if (users[ID]->request_type==REQUEST_MULTIPART2)
    {
      sprintf(temp_string,"\r\n--myboundary\r\nContent-Type: image/jpeg\r\nContent-Length: %ld\r\n\r\n",users[ID]->content_length);
      message(ID,temp_string);
    }

    users[ID]->need_header=NEED_HEADER_NO;
  }

  if (users[ID]->in==-1) return -1;

/*
if (debug==1)
{ printf("sending bytes... %ld left\n",users[ID]->content_length); }
*/

#ifdef NO_INTERLACED_FILE
  while(users[ID]->content_length>0)
#else
  for (c=0; c<CHUNKS_PER_SEND; c++)
#endif
  {
    if (users[ID]->content_length==0) break;

    /* this 1024 thing needs work */
#ifdef WITH_MMAP
    if (users[ID]->video_num>=0)
    {
      if (users[ID]->content_length>BIGGEST_FILE_CHUNK)
      { r=BIGGEST_FILE_CHUNK; }
        else
      { r=users[ID]->content_length; }
    }
      else
#endif
    {
/*
      if (users[ID]->content_length>BIGGEST_FILE_CHUNK)
      { r=fread(temp_string,1,BIGGEST_FILE_CHUNK,users[ID]->in); }
        else
      { r=fread(temp_string,1,users[ID]->content_length,users[ID]->in); }
*/
      if (users[ID]->content_length>BIGGEST_FILE_CHUNK)
      { r=read(users[ID]->in,temp_string,BIGGEST_FILE_CHUNK); }
        else
      { r=read(users[ID]->in,temp_string,users[ID]->content_length); }
    }

    users[ID]->content_length-=r;

#ifdef ENABLE_CGI
    if (r==0 && (users[ID]->mime_type&32768)!=0)
    {
      users[ID]->content_length=0;
      kill_user(ID);
      return 0;
      /* break; */
    }
#endif


    t=0;
    while(t<r)
    {
#ifdef WITH_MMAP
      if (users[ID]->video_num>=0)
      {
        k=send_data(users[ID]->socketid,video[users[ID]->video_num].mem+users[ID]->offset+t,r-t);
      }
        else
#endif
      {
        k=send_data(users[ID]->socketid,temp_string+t,r-t);
      }

#ifdef DEBUG
/* printf("sending k=%d l=%d r=%d\n",k,l,r); */
if (k==-1) printf("k=%d\n",k);
#endif

      if (k==-1)
      {
        kill_user(ID);
        return -1;
      }
      t=t+k;
    }

#ifdef WITH_MMAP
  users[ID]->offset+=r;
#endif

/* printf("here line %d\n",ID); */
  }


/*
if (debug==1)
{ printf("done with 1024 / %ld\n",users[ID]->content_length); }
*/

  if (users[ID]->content_length==0)
  {
    if (users[ID]->video_num==-2)
    {
      /* if (users[ID]->in!=0) fclose(users[ID]->in); */
      if (users[ID]->in!=-1)
      { file_close(ID); }
      users[ID]->video_num=-1;
    }

    if (users[ID]->request_type==REQUEST_SINGLE)
    { users[ID]->state=STATE_IDLE; }
      else
    { users[ID]->need_header=NEED_HEADER_YES; }
  }

  return 0;
}

#ifdef ENABLE_CAPTURE
int send_capture_frame(int ID)
{
char temp_string[BIGGEST_FILE_CHUNK];
// unsigned char *cap_buffer;
int t,k,r,c;

  if (users[ID]->need_header==NEED_HEADER_YES)
  {
    users[ID]->jpeg_ptr=0;
    users[ID]->mime_type=4;

    /* cap_buffer=capture_image(video[users[ID]->video_num].capture_info); */
    users[ID]->content_length=capture_image(video[users[ID]->video_num].capture_info,ID);
#ifdef DEBUG
printf("content-length: %d\n",users[ID]->content_length);
#endif
    if (users[ID]->content_length<0) users[ID]->content_length=0;

/* printf("buffer=%ld len=%d\n",video[users[ID]->video_num].capture_info->buffer,video[users[ID]->video_num].capture_info->buffer_len); */

/*
    if (users[ID]->jpeg_len==0)
    {
// FIXME this should be dynamic instead of wasting memory.. easy to fix
//       also need a way to deallocate this crap

      // users[ID]->jpeg_len=20480;
      users[ID]->jpeg_len=128000;
      users[ID]->jpeg=(unsigned char *)malloc(users[ID]->jpeg_len);
    }
*/

// memset(video[users[ID]->video_num].capture_info->buffer,100,video[users[ID]->video_num].capture_info->buffer_len);

/*
    users[ID]->content_length=
      jpeg_compress(cap_buffer,
                    video[users[ID]->video_num].capture_info->buffer_len,
                    users[ID]->jpeg,
                    users[ID]->jpeg_len,
                    video[users[ID]->video_num].capture_info->width,
                    video[users[ID]->video_num].capture_info->height,
                    3,
                    users[ID]->jpeg_quality);
*/

/* printf("content_length=%d\n",users[ID]->content_length); */


    if (users[ID]->request_type==REQUEST_SINGLE)
    { send_header(ID); }
      else
    if (users[ID]->request_type==REQUEST_MULTIPART)
    { send_header_multipart(ID); }
      else
    if (users[ID]->request_type==REQUEST_MULTIPART2)
    {
      sprintf(temp_string,"\r\n--myboundary\r\nContent-Type: image/jpeg\r\nContent-Length: %ld\r\n\r\n",users[ID]->content_length);
      message(ID,temp_string);
    }

    users[ID]->need_header=NEED_HEADER_NO;
  }

#ifdef NO_INTERLACED_FILE
  while(users[ID]->jpeg_ptr<users[ID]->content_length)
#else
  for (c=0; c<CHUNKS_PER_SEND; c++)
#endif
  {
    /* this BIGGEST_FILE_CHUNK thing needs work */

    if (users[ID]->content_length-users[ID]->jpeg_ptr>BIGGEST_FILE_CHUNK)
    { r=BIGGEST_FILE_CHUNK; }
      else
    { r=users[ID]->content_length-users[ID]->jpeg_ptr; }

    t=0;
    while(t<r)
    {
      k=send_data(users[ID]->socketid,(char *)(users[ID]->jpeg+users[ID]->jpeg_ptr+t),r-t);
      if (k==-1)  // FIXME
      {
        kill_user(ID);
        return -1;
      }
      t=t+k;
    }

    users[ID]->jpeg_ptr+=t;
/* printf("%d %d\n",users[ID]->jpeg_ptr,users[ID]->content_length); */
    if (users[ID]->jpeg_ptr==users[ID]->content_length) break;
  }

  if (users[ID]->jpeg_ptr==users[ID]->content_length)
  {
    if (users[ID]->request_type==REQUEST_SINGLE)
    { users[ID]->state=STATE_IDLE; }
      else
    { users[ID]->need_header=NEED_HEADER_YES; }
  }

  return 0;
}
#endif

int buffered_read(int ID)
{
int r,t=0;

  while(1)
  {
    errno=0;
    if (users[ID]->sptr>=users[ID]->slen)
    {
      users[ID]->slen=recv(users[ID]->socketid,(char *)users[ID]->myString,BUFFER_SIZE-2,0);
      users[ID]->sptr=0;

/* printf("recv=%d\n",users[ID]->slen); */

      if (users[ID]->slen==0) return -1;
      if (users[ID]->slen<0)
      {
/* FIXME
        if (users[ID]->inuse==1)
         if (errno==EINTR || errno==EAGAIN) 
         { errno=0; return 0; }
 */ 
#ifdef DEBUG
        if (debug==1)
        { printf("Lost Connection (NULL): %d\n",ID); fflush(stdout); }
#endif

        users[ID]->buffer_ptr=0; 
        users[ID]->slen=0;
        return -1;
      }
    }

    // FIXME - do we need this?
    users[ID]->myString[users[ID]->slen]=0;

    for (r=users[ID]->sptr; r<users[ID]->slen; r++)
    { 
      if ((users[ID]->myString[r]=='\n' || users[ID]->myString[r]=='\r') || users[ID]->myString[r]==0)
      {
        if (r+1<users[ID]->slen)
        {
          if (users[ID]->myString[r]=='\n')
          {
            if (users[ID]->myString[r+1]=='\r') { r++; }
          }
            else
          if (users[ID]->myString[r]=='\r')
          {
            if (users[ID]->myString[r+1]=='\n') { r++; }
          }
        }

        t=-1;
        break;
      }

      if (users[ID]->myString[r]==127 || users[ID]->myString[r]==8) 
      { if (users[ID]->buffer_ptr>0) users[ID]->buffer_ptr--; }
        else
      { if (users[ID]->myString[r]>=32 && users[ID]->myString[r]<254) users[ID]->outString[users[ID]->buffer_ptr++]=users[ID]->myString[r]; }

      t=1;
    }

    users[ID]->sptr=r+1;

    if ((t==-1 && users[ID]->buffer_ptr>0) || users[ID]->buffer_ptr>BUFFER_SIZE-2) 
    { 
      users[ID]->outString[users[ID]->buffer_ptr]=0;
      users[ID]->buffer_ptr=0;
      if (users[ID]->myString[users[ID]->sptr]=='\n' || 
          users[ID]->myString[users[ID]->sptr]=='\r')
      { users[ID]->sptr++; }

      return 1; 
    }
      else
    if (t==-1 && users[ID]->buffer_ptr==0)
    { return 2; }
      else
    { return 0; }
  }
}


