#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#ifndef WINDOWS
#include <sys/types.h>
#include <sys/time.h>
#include <pwd.h>
#include <grp.h>
#include <dlfcn.h>
#endif
#include <errno.h>
#include <time.h>
#include <ctype.h>
#ifdef WINDOWS
#include <windows.h>
#endif
#ifdef WITH_MMAP
#ifndef WINDOWS
#include <sys/mman.h>
#endif
#endif

#include "globals.h"
#include "functions.h"
#include "avi.h"

/*

This code is copyright 2004-2007 by Michael Kohn
You may not use any of this program in your own programs
without permission from Michael Kohn <mike@mikekohn.net>

*/

#ifdef ENABLE_CAPTURE
static void parse_size(char *s, int *width, int *height)
{
int ptr,r,k;

  ptr=0;
  r=0;
  k=0;

  while(1)
  {
    if (s[ptr]==0 && k==0) break;
    if (s[ptr]=='x' || s[ptr]==',' || s[ptr]==0)
    {
      s[ptr++]=0;
      if (k==0)
      { *width=atoi(s+r); }
        else
      { *height=atoi(s+r); }

      k++;
      if (k==2) break;
      r=ptr;
      continue;
    }

    ptr++;
  }
}
#endif

static int gettoken(FILE *in, char *token, int token_len)
{
int ch,ptr;

  ptr=0;
  token[0]=0;

  while(1)
  {
    ch=getc(in);
    if (ch=='#' && ptr==0)
    {
      while(1)
      {
        ch=getc(in);
        if (ch=='\n' || ch=='\r' || ch==EOF) break;
      }

      continue;
    }

    if (ch==';' || ch=='{' || ch=='}')
    {
      if (ptr==0)
      {
        token[ptr++]=ch;
        break;
      }
      ungetc(ch,in);
      break;
    }

    if (ch=='\n' || ch=='\r' || ch==' ' || ch=='\t' || ch==EOF)
    {
      if (ptr!=0) break;
      if (ch==EOF) return -1;
      continue;
    }

    token[ptr++]=ch;
    if (ptr>=token_len-1) break;
  }

  token[ptr]=0;

  return 0;
}

static int setup_file(char *filename)
{
FILE *in;

  in=fopen(filename,"rb");

  if (in==0)
  {
    printf("Cannot open AVI file: %s\n",filename);
    return -1;
  }

  printf("Indexing AVI file %d: '%s'\n",video_count,filename);

  memset(&video[video_count],0,sizeof(struct video_t));

  video[video_count].filename=malloc(strlen(filename)+1);

  strcpy(video[video_count].filename,filename);
  parse_riff(in,&video[video_count]);

  /* gettimeofday(&video[video_count].tv_start,DST_NONE); */
  gettimeofday(&video[video_count].tv_start,0);

  video[video_count].tv_total.tv_sec=video[video_count].total_frames/video[video_count].fps;
  video[video_count].tv_total.tv_usec=(int)((((double)video[video_count].total_frames/(double)video[video_count].fps)-(double)video[video_count].tv_total.tv_sec)*(double)1000000);

if (debug==1)
{
  printf("fps=%d total_frames=%d sec=%d usec=%d\n",video[video_count].fps,video[video_count].total_frames,(int)video[video_count].tv_total.tv_sec,(int)video[video_count].tv_total.tv_usec);
}

#ifdef WITH_MMAP
  fseek(in,0,SEEK_END);
  video[video_count].file_len=ftell(in);
#ifndef WINDOWS
  video[video_count].fd=open(video[video_count].filename,O_RDONLY);
  video[video_count].mem=(unsigned char *)mmap(NULL,video[video_count].file_len,PROT_READ,MAP_SHARED,video[video_count].fd,0);
#else
  video[video_count].fd=CreateFile(video[video_count].filename,FILE_READ_DATA,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_READONLY,NULL);
  // Security.nLength=sizeof(Security);
  // Security.lpSecurityDescriptor=NULL;
  video[video_count].mem_handle=CreateFileMapping(video[video_count].fd,NULL,PAGE_READONLY,0,video[video_count].file_len,NULL);
  video[video_count].mem=(unsigned char *)MapViewOfFile(video[video_count].mem_handle,FILE_MAP_READ,0,0,video[video_count].file_len);
#endif
#endif

  video_count++;
  fclose(in);

  return 0;
}

#ifdef ENABLE_CGI
static int parse_handler(FILE *in)
{
struct cgi_handler_t *curr_handler,*temp_handler;
char token[2048];

  curr_handler=(struct cgi_handler_t *)malloc(sizeof(struct cgi_handler_t));

  if (cgi_handler==0)
  { cgi_handler=curr_handler; }
    else
  {
    temp_handler=cgi_handler;
    while(temp_handler->next_handler!=0)
    { temp_handler=temp_handler->next_handler; }
    temp_handler->next_handler=curr_handler;
  }

  memset(curr_handler,0,sizeof(struct cgi_handler_t));

  gettoken(in,token,2048);
  curr_handler->extension=(char *)malloc(strlen(token)+1);
  strcpy(curr_handler->extension,token);

  gettoken(in,token,2048);
  if (strcmp(token,";")==0) return 0;
  if (strcmp(token,"{")!=0) return -1;

  while(1)
  {
    if (gettoken(in,token,2048)!=0)
    {
      printf("Parse error in alias, expected '}'\n");
      return -1;
    }

    if (strcmp(token,"}")==0) break;

    if (strcasecmp(token,"application")==0)
    {
      gettoken(in,token,2048);
      curr_handler->application=(char *)malloc(strlen(token)+1);
      strcpy(curr_handler->application,token);
    }
  }

  return 0;
}
#endif

#ifdef ENABLE_PLUGINS

static int parse_plugin(FILE *in)
{
struct plugin_t *curr_plugin,*temp_plugin;
naken_plugin_init_t naken_plugin_init = 0;
char token[2048];

  curr_plugin=(struct plugin_t *)malloc(sizeof(struct plugin_t));

  if (plugin==0)
  { plugin=curr_plugin; }
    else
  {
    temp_plugin=plugin;
    while(temp_plugin->next_plugin!=0)
    { temp_plugin=temp_plugin->next_plugin; }
    temp_plugin->next_plugin=curr_plugin;
  }

  memset(curr_plugin,0,sizeof(struct plugin_t));

  gettoken(in,token,2048);
  curr_plugin->alias=(char *)malloc(strlen(token)+1);
  strcpy(curr_plugin->alias,token);
  curr_plugin->alias_len=strlen(token);

  gettoken(in,token,2048);
  if (strcmp(token,"{")!=0) return -1;

  while(1)
  {
    if (gettoken(in,token,2048)!=0)
    {
      printf("Parse error in plugin, expected '}'\n");
      return -1;
    }

    if (strcmp(token,"}")==0) break;

    if (strcasecmp(token,"library")==0)
    {
      gettoken(in,token,2048);
      curr_plugin->dlhandle=dlopen(token,RTLD_LAZY);
      if (curr_plugin->dlhandle==NULL)
      {
        printf("Error opening up library %s\n",token);
        continue;
      }
      naken_plugin_init=(naken_plugin_init_t)dlsym(curr_plugin->dlhandle,"naken_plugin_init");
      curr_plugin->naken_plugin_get=(naken_plugin_get_t)dlsym(curr_plugin->dlhandle,"naken_plugin_get");
      curr_plugin->naken_plugin_post=(naken_plugin_post_t)dlsym(curr_plugin->dlhandle,"naken_plugin_post");
    }
  }

  if (naken_plugin_init==0 ||
      curr_plugin->naken_plugin_get==0 ||
      curr_plugin->naken_plugin_post==0)
  {
    printf("Error loading plugin\n");
    curr_plugin->alias[0]=0;
  }
    else
  {
    naken_plugin_init();
  }

  return 0;
}

#endif

static int parse_alias(FILE *in)
{
struct alias_t *curr_alias,*temp_alias;
char token[2048];
int r;

  curr_alias=(struct alias_t *)malloc(sizeof(struct alias_t));

  if (alias==0)
  { alias=curr_alias; }
    else
  {
    temp_alias=alias;
    while (temp_alias->next_alias!=0) temp_alias=temp_alias->next_alias;
    temp_alias->next_alias=curr_alias;
  }

  memset(curr_alias,0,sizeof(struct alias_t));

  gettoken(in,token,2048);
  curr_alias->url_len=strlen(token);
  curr_alias->url=(char *)malloc(strlen(token)+1);
  strcpy(curr_alias->url,token);

  gettoken(in,token,2048);
  if (strcmp(token,"{")!=0)
  {
    printf("Parse error in alias, expected '{' and got '%s'\n",token);
    return -1;
  }

  while(1)
  {
    if (gettoken(in,token,2048)!=0)
    {
      printf("Parse error in alias, expected '}' and got '%s'\n",token);
      return -1;
    }

    if (strcmp(token,"}")==0) break;

    if (strcasecmp(token,"port")==0)
    {
      gettoken(in,token,2048);
      r=conv_num(token);

      if (r!=-1)
      { curr_alias->port=r; }
        else
      {
        curr_alias->port_param=(char *)malloc(strlen(token)+1);
        strcpy(curr_alias->port_param,token);
      }
    }
      else
    if (strcasecmp(token,"type")==0)
    {
      gettoken(in,token,2048);
      if (strcasecmp(token,"single")==0) curr_alias->type=0;
        else
      if (strcasecmp(token,"stream")==0) curr_alias->type=1;
    }
      else
    if (strcasecmp(token,"size")==0 && curr_alias->size_param==0)
    {
      gettoken(in,token,2048);
      curr_alias->size_param=(char *)malloc(strlen(token)+1);
      strcpy(curr_alias->size_param,token);
    }
      else
    if (strcasecmp(token,"comp")==0 && curr_alias->comp_param==0)
    {
      gettoken(in,token,2048);
      curr_alias->comp_param=(char *)malloc(strlen(token)+1);
      strcpy(curr_alias->comp_param,token);
    }
      else
    if (strcasecmp(token,"frame_rate")==0 && curr_alias->fps_param==0)
    {
      gettoken(in,token,2048);
      curr_alias->fps_param=(char *)malloc(strlen(token)+1);
      strcpy(curr_alias->fps_param,token);
    }
      else
    {
      printf("Error in conf '%s'\n",token);
      return -1;
    }
  }

  return 0;
}

#ifdef ENABLE_CAPTURE
static int parse_capture(FILE *in)
{
char token[1024];
char value[1024];
char video_dev[1024];
int i;

  gettoken(in,video_dev,1024);

  gettoken(in,token,1024);
  if (strcmp(token,"{")!=0)
  {
    printf("Parse error in alias, expected '{' and got '%s'\n",token);
    return -1;
  }

  memset(&video[video_count],0,sizeof(struct video_t));
  video[video_count].capture_info=(struct capture_info_t *)malloc(sizeof(struct capture_info_t));

  video[video_count].capture_info->width=352;
  video[video_count].capture_info->height=240;
  video[video_count].capture_info->channel=-1;
#ifdef V4L
  video[video_count].capture_info->format=V4L2_STD_NTSC_M;
#endif

  while(1)
  {
    if (gettoken(in,token,1024)==-1) return -1;

    if (strcmp(token,"}")==0) break;

    gettoken(in,value,1024);

    if (strcmp(token,"size")==0)
    {
      parse_size(value,&video[video_count].capture_info->width,&video[video_count].capture_info->height);
    }
      else
    if (strcmp(token,"max_fps")==0)
    { video[video_count].capture_info->max_fps=atoi(value); }
      else
    if (strcmp(token,"channel")==0)
    { video[video_count].capture_info->channel=atoi(value); }
      else
    if (strcmp(token,"format")==0)
    {
#ifdef V4L
      if (strcasecmp(value,"pal")==0)
      { video[video_count].capture_info->format=V4L2_STD_PAL_M; }
        else
      if (strcasecmp(value,"ntsc")==0)
      { video[video_count].capture_info->format=V4L2_STD_NTSC_M; }
        else
      { video[video_count].capture_info->format=atoi(value); }
#endif
    }
  }

  printf("%s  %dx%d  max_fps=%d\n",video_dev,video[video_count].capture_info->width,video[video_count].capture_info->height,video[video_count].capture_info->max_fps);

  i=open_capture(video[video_count].capture_info,video_dev);

  if (i==0)
  { video_count++; }
    else
  {
    printf("Couldn't open capture for device %s (error %d)\n",video_dev,i);
    free(video[video_count].capture_info);
  }

  return 0;
}
#endif

#ifndef WINDOWS
void set_runas()
{
struct passwd *pw;
struct group *gr;

  if (runas_user!=0)
  {
    pw=getpwnam(runas_user);

    if (pw==0)
    { printf("No such user %s\n",runas_user); }
      else
    {
      if (runas_group!=0)
      {
        gr=getgrnam(runas_group);
        if (gr==0)
        { printf("No such group %s\n",runas_group); }
          else
        if (setgid(gr->gr_gid)!=0)
        { printf("Couldn't set group %s\n",runas_group); }

        free(runas_group);
        runas_group=0;
      }

      if (setuid(pw->pw_uid)!=0)
      { printf("Couldn't set user to %s\n",runas_user); }
    }

    free(runas_user);
    runas_user=0;
  }
}

#endif

#if !defined(ENABLE_CGI) || !defined(ENABLE_PLUGINS) || !defined(ENABLE_CAPTURE)
static void skip_block(FILE *in)
{
char token[1024];
int brace_count=0;

  while(gettoken(in,token,1024)==0)
  {
    if (strcmp(token,"{")==0) brace_count++;
      else
    if (strcmp(token,"}")==0)
    {
      brace_count--;
      if (brace_count==0) return;
    }
      else
    if (strcmp(token,";")==0 && brace_count==0) return;
  }
}
#endif

void read_config(char *config_dir)
{
FILE *in;
char token[1024];
char username[512];
char password[256];
#ifndef WINDOWS
int ch;
#endif

  username[0]=0;
  password[0]=0;

  in=fopen(config_dir,"rb");

  if (in==0)
  {
    printf("No config file found.  Going with defaults.\n");
    return;
  }

  while(gettoken(in,token,1024)==0)
  {
    if (strcasecmp(token,"filename")==0)
    {
      gettoken(in,token,1024);
      setup_file(token);
    }
      else
    if (strcasecmp(token,"port")==0)
    {
      gettoken(in,token,1024);
      port=atoi(token);
    }
      else
    if (strcasecmp(token,"maxconn")==0)
    {
      gettoken(in,token,1024);
      maxconn=atoi(token);
    }
      else
    if (strcasecmp(token,"minconn")==0)
    {
      gettoken(in,token,1024);
      minconn=atoi(token);
    }
      else
    if (strcasecmp(token,"max_idle_time")==0)
    {
      gettoken(in,token,1024);
      max_idle_time=atoi(token);
    }
      else
    if (strcasecmp(token,"htdocs_dir")==0)
    {
      gettoken(in,token,1024);
      htdocs_dir=(char *)malloc(strlen(token)+1);
      strcpy(htdocs_dir,token);
    }
      else
    if (strcasecmp(token,"username")==0)
    {
      gettoken(in,username,512);
    }
      else
    if (strcasecmp(token,"password")==0)
    {
      gettoken(in,password,512);
    }
      else
    if (strcasecmp(token,"runas")==0)
    {
      gettoken(in,token,1024);

#ifdef WINDOWS
      printf("runas is not supported on Windows\n");
#else
      ch=0;
      while(token[ch]!=':' && token[ch]!=0) ch++;

      if (token[ch]==':')
      { token[ch++]=0; }
        else
      { ch++; token[ch]=0; }

      if (runas_user==0)
      {
        runas_user=(char *)malloc(strlen(token)+1);
        strcpy(runas_user,token);
      }

      if (token[ch]!=0)
      {
        // gettoken(in,token,128);
        if (runas_group==0)
        {
          runas_group=(char *)malloc(strlen(token+ch)+1);
          strcpy(runas_group,token+ch);
        }
      }
#endif
    }
      else
    if (strcasecmp(token,"jpeg_quality")==0)
    {
      gettoken(in,token,1024);
      jpeg_quality=atoi(token);
    }
      else
    if (strcasecmp(token,"frame_rate")==0)
    {
      gettoken(in,token,1024);
      frame_rate=atoi(token);
    }
      else
    if (strcasecmp(token,"alias")==0)
    {
      parse_alias(in);
    }
      else
    if (strcasecmp(token,"cgi_handler")==0)
    {
#ifdef ENABLE_CGI
      parse_handler(in);
#else
      printf("CGI support was not compiled in.\n");
      skip_block(in);
#endif
    }
      else
    if (strcasecmp(token,"plugin")==0)
    {
#ifdef ENABLE_PLUGINS
      parse_plugin(in);
#else
      printf("Plugin support was not compiled in.\n");
      skip_block(in);
#endif
    }
      else
    if (strcasecmp(token,"capture")==0)
    {
#ifdef ENABLE_CAPTURE
      parse_capture(in);
#else
      printf("Capture support was not compiled in.\n");
      skip_block(in);
#endif
    }
      else
    {
      printf("Config file syntax error: %s\n",token);
      break;
    }
  }

  fclose(in);

  if (username[0]!=0 || password[0]!=0)
  {
    strcat(username,":");
    strcat(username,password);

    if (base64_encode(username)!=0) user_pass_64[0]=0;
  }

}

