#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <windows.h>
// #include <winuser.h>
#include <vfw.h>

#include "capture.h"
#include "jpeg_compress.h"

#define WM_CAP_START WM_USER
#define WM_CAP_SET_CALLBACK_FRAME WM_CAP_START+5
#define WM_CAP_DRIVER_CONNECT WM_CAP_START+10 
#define WM_CAP_DRIVER_DISCONNECT WM_CAP_START+11 
#define WM_CAP_SET_VIDEOFORMAT WM_CAP_START+45 
#define WM_CAP_GRAB_FRAME WM_CAP_START+60 
#define WM_CAP_STOP WM_CAP_START+68

#define MAX_CAPTURE_DEVICES 128

/*
static struct capture_info_ptr_t
{
  struct capture_info_t *capture_info;
} capture_info_ptr[MAX_CAPTURE_DEVICES];
*/

static struct capture_info_t *capture_info_ptr[MAX_CAPTURE_DEVICES];

static int vfw_count=0;

typedef struct
{
  unsigned char *lpData;
  DWORD dwBufferLength;
} *LPVIDEOHDR;

LRESULT CALLBACK capVideoStreamCallback(HWND hWnd, LPVIDEOHDR lpVHdr)
{
int i;

printf("In callback\n");
fflush(stdout);

  for (i=0; i<vfw_count; i++)
  {
    if (capture_info_ptr[i]==NULL) return -1;
    if (capture_info_ptr[i]->hWnd) break;
  }
  if (i==vfw_count) return -2;

#ifdef DEBUG
printf("In callback: Copy\n");
fflush(stdout);
#endif

  memcpy(capture_info_ptr[i]->buffer,lpVHdr->lpData,lpVHdr->dwBufferLength);
  capture_info_ptr[i]->callback_wait=0;
  capture_info_ptr[i]->jpeg_len=lpVHdr->dwBufferLength;

#ifdef DEBUG
printf("exit callback\n");
fflush(stdout);
#endif

  return 0;
}

int open_capture(struct capture_info_t *capture_info, char *dev_name)
{
int i;

  if (vfw_count==0)
  {
    for (i=0; i<MAX_CAPTURE_DEVICES; i++) capture_info_ptr[i]=NULL;
  }

  // capture_info_ptr[vfw_count++]=(struct capture_info_t *)malloc(sizeof(struct capture_info_t *));
  capture_info_ptr[vfw_count++]=capture_info;

  capture_info->hWnd=capCreateCaptureWindow("Naken Web",
                                            0,
                                            capture_info->width,
                                            capture_info->height,
                                            capture_info->width,
                                            capture_info->height,
                                            NULL,
                                            0);

  if (capture_info->hWnd==0) return -1;

  // capDriverConnect(capture_info->hWnd,atoi(dev_name));
  // i=SendMessage(capture_info->hWnd,WM_CAP_DRIVER_CONNECT,atoi(dev_name),0);
// printf("blah: %d %d\n",i,FALSE);

  if (!SendMessage(capture_info->hWnd,WM_CAP_DRIVER_CONNECT,atoi(dev_name),0))
  { return -2; }

  memset(&capture_info->bmp,0,sizeof(BITMAPINFO));
  capture_info->bmp.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
  capture_info->bmp.bmiHeader.biBitCount=24;
  capture_info->bmp.bmiHeader.biPlanes=1;
  capture_info->bmp.bmiHeader.biWidth=capture_info->width;
  capture_info->bmp.bmiHeader.biHeight=capture_info->height;
  capture_info->bmp.bmiHeader.biSizeImage=capture_info->width*capture_info->height*3;
  capture_info->bmp.bmiHeader.biCompression=BI_RGB;

  if (!SendMessage(capture_info->hWnd,WM_CAP_SET_VIDEOFORMAT,sizeof(BITMAPINFO),(LPARAM)&capture_info->bmp))
  { return -3; }

  capture_info->buffer_len=capture_info->width*capture_info->height*3;
  capture_info->buffer=(unsigned char *)malloc(capture_info->buffer_len);

  // if (!SendMessage(capture_info->hWnd,WM_CAP_SET_CALLBACK_FRAME,0,(LPARAM)0))
  if (!SendMessage(capture_info->hWnd,WM_CAP_SET_CALLBACK_FRAME,0,(LPARAM)capVideoStreamCallback))
  { return -4; }

  return 0;
}

int capture_image(struct capture_info_t *capture_info, int ID)
{
int count;

  capture_info->callback_wait=1;

#ifdef DEBUG
printf("capture_image: enter %d  %x\n",(int)capture_info->hWnd,WM_CAP_GRAB_FRAME);
fflush(stdout);
#endif

  if (!SendMessage(capture_info->hWnd,WM_CAP_GRAB_FRAME,0,0))
  {
#ifdef DEBUG
printf("capture_image: grab error\n");
fflush(stdout);
#endif
    memset(capture_info->buffer,0,capture_info->buffer_len);
  }

#ifdef DEBUG
printf("capture_image: waiting\n");
fflush(stdout);
#endif

  count=0;
  while(capture_info->callback_wait==1)
  {
    Sleep(5);
    count++;
    if (count==100) break;
    count++;
  }
#ifdef DEBUG
printf("capture_image: exit %d\n",count);
fflush(stdout);
#endif

  return capture_info->buffer;
}

int close_capture(struct capture_info_t *capture_info)
{
  SendMessage(capture_info->hWnd,WM_CAP_SET_CALLBACK_FRAME,0,0);
  SendMessage(capture_info->hWnd,WM_CAP_STOP,0,0);

  free(capture_info->buffer);
  if (capture_info->picture!=0) free(capture_info->picture);

  return 0;
}



